package server

import (
	"errors"
	odysseus "github.com/odysseus/odysseus-protocol/gen/proto/go/base/v1"
	omanager "github.com/odysseus/odysseus-protocol/gen/proto/go/nodemanager/v2"
	log "github.com/sirupsen/logrus"
	"sync"
	"time"
)

type dispatchTask struct {
	create     time.Time
	resultTime time.Time
	worker     *Worker
	task       *odysseus.TaskContent
	status     TaskStatus
	ack        chan interface{}
	errCh      chan error
	mux        sync.Mutex
	result     chan *omanager.SubmitTaskResult
}

func newDispatchTask(w *Worker, task *odysseus.TaskContent) *dispatchTask {
	return &dispatchTask{
		create: time.Now(),
		worker: w,
		task:   task,
		status: TASK_CREATE,
		ack:    make(chan interface{}, 1),
		errCh:  make(chan error, 1),
		result: make(chan *omanager.SubmitTaskResult, 1),
	}
}
func (d *dispatchTask) dispatched(wm *WorkerManager) {
	tm := time.NewTicker(time.Second * 5)
	defer tm.Stop()
	d.setStatus(TASK_WAIT_ACK)
	t1 := time.Now()

	select {
	case ackMsg, ok := <-d.ack:
		if !ok {
			d.errCh <- errors.New("ack channel closed")
		} else {
			msg := ackMsg.(*omanager.SubmitTaskAck)
			log.WithFields(log.Fields{
				"canExecute":   msg.CanExecute,
				"task-id":      msg.TaskId,
				"bootup-time":  msg.BootUpTime,
				"queue-time":   msg.QueueWaitTime,
				"execute-time": msg.ExecuteTime,
				"ackcost":      time.Since(t1).Milliseconds(),
			}).Debug("got ack message")
			if !msg.CanExecute {
				d.errCh <- errors.New("worker can't execute task")
			} else {
				d.setStatus(TASK_ACKED)
				d.errCh <- nil
			}
		}
	case <-tm.C:
		d.errCh <- errors.New("ack timeout")
	}

	if d.status == TASK_ACKED {
		// wait timeout or got result
		d.finalize(wm)
	}
	return

}
func (d *dispatchTask) setAck(msg *omanager.SubmitTaskAck) {
	select {
	case d.ack <- msg:
	default:
	}
}

func (d *dispatchTask) setResult(result *omanager.SubmitTaskResult) {
	select {
	case d.result <- result:
	default:
	}
}

func (d *dispatchTask) setStatus(status TaskStatus) {
	d.mux.Lock()
	d.status = status
	d.mux.Unlock()
}

func (d *dispatchTask) finalize(wm *WorkerManager) {
	maxExec := d.task.TaskMaxExecTime
	if maxExec <= 0 {
		maxExec = 300 // set default to 5min.
	}
	l := log.WithField("task-id", d.task.TaskId)
	ticker := time.NewTicker(time.Second * time.Duration(maxExec))
	defer ticker.Stop()
	var result *omanager.SubmitTaskResult
	select {
	case <-ticker.C:
		d.setStatus(TASK_TIMEOUT)
		result = &omanager.SubmitTaskResult{
			TaskId:              d.task.TaskId,
			IsSuccessed:         false,
			TaskExecuteDuration: uint64(maxExec * 1000 * 1000), // 微秒
			TaskExecuteError:    ErrExecuteTimeout.Error(),
		}

	case r := <-d.result:
		d.setStatus(TASK_FINISHED)
		result = r
	}
	d.resultTime = time.Now()
	l.WithFields(log.Fields{
		"totaltime": d.resultTime.Sub(d.create).Milliseconds(),
		"status":    d.status,
	}).Debug("task finished")

	task := d.task
	if task.TaskKind != odysseus.TaskKind_StandardTask && d.worker.online == true {

	}

	_, err := wm.taskResult(d.worker, task, result)
	if err != nil {
		log.WithError(err).Error("task result failed")
	}

	if task.TaskKind != odysseus.TaskKind_StandardTask {
		wm.Payment(task)
	}

}
